package server.threads;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.util.Vector;

import server.data.FileElement;
import server.data.UtilMethods;
import auxdata.ConInfo;


public class TCPUploadHandlerThread extends Thread {

	private Vector<FileElement> fileStore = null;

	private Socket socket = null;

	private int tryLimit = 10;

	private int tryCount = 0;

	public TCPUploadHandlerThread(Socket socket, Vector<FileElement> fileStore) {
		super("UploadHandlerThread");
		this.socket = socket;
		this.fileStore = fileStore;
		System.out.println("fileStore size: " + fileStore.size());
	}

	public void run () {
		try {
			byte[] bytedFileSize = new byte[4]; // An integer
			int fileSize;
			byte[] fileBuf;
			int bytesRead;
			InputStream is;
			OutputStream os;

			// Get filesize of file to be received
			is = socket.getInputStream();
			BufferedInputStream bis = new BufferedInputStream(is);
			bytesRead = bis.read(bytedFileSize,0,bytedFileSize.length);

			// Take note of the file size to be received
			fileSize = UtilMethods.bytesToInt(bytedFileSize, 0);
			// Make a buffer large enough for the file.
			fileBuf = new byte[fileSize];

			// Notifiy sender of filesize received
			os = socket.getOutputStream();
			os.write(ConInfo.FILESIZERECEIVEDOK.getBytes());

			// Receive the file
			System.out.println("Receving " + fileSize + " bytes");
			bytesRead = bis.read(fileBuf,0,fileBuf.length);
			System.out.println("Received " + bytesRead + " bytes");

			// Split file into constituent parts
			
			byte[][] fileParts = splitFile(fileBuf);

			FileElement part1 = new FileElement(
					fileStore.size(),
					FileElement.FIRSTPART, 
					(fileSize % 2 == 0), 
					fileParts[0]);
			FileElement part2 = new FileElement(
					fileStore.size(), 
					FileElement.LASTPART, 
					(fileSize % 2 == 0), 
					fileParts[1]);
			FileElement xorSum = new FileElement(
					fileStore.size(), 
					FileElement.XORSUM, 
					(fileSize % 2 == 0), 
					fileParts[2]);
			
			// Save the xorSum in the locale fileStore
			fileStore.add(xorSum);
			
			// Distribute the two fileparts on other servers
			
			InetAddress server1 = ConInfo.getTheTwoOtherServersIpAdresses()[0];
			InetAddress server2 = ConInfo.getTheTwoOtherServersIpAdresses()[1];
			
			int sendOne = UDPSendFile(part1.getBytes(), server1);
			int sendTwo = UDPSendFile(part2.getBytes(), server2);

			// Return a ticket number to the client
			if (sendOne > 0 || sendTwo > 0) {
				os.write(UtilMethods.intToBytes(fileStore.indexOf(xorSum))); // Writes exactly 4 bytes	
			} else { // File has not been distributed
				os.write(UtilMethods.intToBytes(-1)); // Writes exactly 4 bytes
			}
			

			is.close();
			os.close();
			socket.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param file
	 * 
	 * 
	 * @return
	 * 
	 * First half of the file in byte[0],
	 * second half of the file in byte[1],
	 * XOR sum of the file in byte[2].
	 */
	private byte[][] splitFile (byte[] file) {

		byte[][] result = null;

		if (file.length % 2 == 0) {
			// File can be split into two equal sizes

			result = new byte[3][file.length / 2];

			for (int i = 0; i < file.length; i++) {
				if (i < file.length / 2) {
					result[0][i] = file[i]; 
				} else {
					result[1][i - (file.length / 2)] = file[i];
				}
			}

			result[2] = xorSUM(result[0], result[1]);

		} else {
			// File can be split into length / 2 and length / 2 + 1
			// But for convenience we extend the smallest array, so 
			// we end up with two arrays of length + 1.

			result = new byte[3][(file.length + 1) / 2 ];

			for (int i = 0; i < file.length + 1; i++) {

				if (i < file.length) {
					if (i < (file.length + 1) / 2) {
						result[0][i] = file[i];
					} else {
						result[1][i - ((file.length + 1) / 2)] = file[i];
					}
				} else {
					result[1][i] = 0; // This byte does not exist in the original file	
				}

			}

			result[2] = xorSUM(result[0], result[1]);

		}

		return result;
	}

	/**
	 * @param a
	 * @param b
	 * @return
	 * @requires a.length == b.length
	 */
	private byte[] xorSUM (byte [] a, byte [] b) {
		byte[] result;

		result = new byte[a.length];
		for (int i = 0; i < a.length; i++) {
			result[i] = (byte) (a[i] ^ b[i]);
		}
		return result;
	}
	
	private int UDPSendFile (byte[] data, InetAddress ia) {
		
		DatagramPacket sendPacket = null;
		DatagramPacket receivePacket = null;
		
		byte[] sendData = null;
		
		String requestCode = null;
		String returnCode = null;
		
		byte[] protocolBuffer = new byte[16];
		
		DatagramSocket ds = null;
		
		try {
			ds = new DatagramSocket();
			ds.setSoTimeout(ConInfo.UDPANSWERTIMEOUT); // Stop blocking on receive() after timeout seconds
			
			// Request upload to the server:
			requestCode = new String(ConInfo.SERVERRQFILEUPLOAD);
			sendPacket = new DatagramPacket(
					requestCode.getBytes(), requestCode.getBytes().length,
					ia, ConInfo.UDPLISTENPORT);
			ds.send(sendPacket);
			
			// wait on an acknowledgement from server			
			while (tryCount++ < tryLimit) {
				receivePacket = new DatagramPacket(protocolBuffer, protocolBuffer.length);
				ds.receive(receivePacket);
				returnCode = new String(receivePacket.getData(), 0, receivePacket.getLength());
				if (returnCode.equals(ConInfo.SERVERRQFILEUPLOADOK)) {
					break; // We got the ok, lets go
				} else { // does not compute, hit me again
					sendPacket = new DatagramPacket(
							ConInfo.NAK.getBytes(), ConInfo.NAK.length(),
							ia, ConInfo.UDPRECEIVEPORT);
					ds.send(sendPacket);
				}
				
				if (tryCount == tryLimit) {
					System.err.println("Failed to receive an ok from server on filesize");
					return -1;
				}
				
			}
			
			// send the filesize to be accepted
			byte[] bytedFileSize = UtilMethods.intToBytes(data.length);
			sendPacket = new DatagramPacket(
					bytedFileSize, bytedFileSize.length,
					ia, ConInfo.UDPRECEIVEPORT);
			ds.send(sendPacket);
			
			// wait on an acknowledgement from server
			while (tryCount++ < tryLimit) {
				receivePacket = new DatagramPacket(protocolBuffer, protocolBuffer.length);
				ds.receive(receivePacket);
				returnCode = new String(receivePacket.getData(), 0, receivePacket.getLength());
				if (returnCode.equals(ConInfo.FILESIZERECEIVEDOK)) {
					break; // We got the ok, lets go
				} else { // does not compute, hit me again
					sendPacket = new DatagramPacket(
							ConInfo.NAK.getBytes(), ConInfo.NAK.length(),
							ia, ConInfo.UDPRECEIVEPORT);
					ds.send(sendPacket);
				}
				
				if (tryCount == tryLimit) {
					System.err.println("Failed to receive an ok from server on filesize");
					return -1;
				}
				
			}
			
			// Now the fun begins, filedata, FIRE AWAY
			
			int blockCount = UtilMethods.getBlocksNeeded(data.length);
			for (int i = 0; i < blockCount; i++) {
				sendData = getBlock(data, i);
				sendPacket = new DatagramPacket(
						sendData, sendData.length,
						ia, ConInfo.UDPLISTENPORT);
				ds.send(sendPacket);
			}
			
			// all data sent, wait for resend request of packets
			// keep resending packets until a sequence number < 0
			// is received
			int blockMissing;
			do {
				receivePacket = new DatagramPacket(
						protocolBuffer, protocolBuffer.length);
				ds.receive(receivePacket);
				if ((blockMissing = UtilMethods.bytesToInt(receivePacket.getData(), 0)) > 0) {
					sendData = getBlock(data, blockMissing);
					sendPacket = new DatagramPacket(
							sendData, sendData.length,
							ia, ConInfo.UDPLISTENPORT);
					ds.send(sendPacket);
				}
			} while (blockMissing > 0);
						
		} catch (SocketException e) {
			System.err.println("Timed out waiting for an ok from server: " + receivePacket.getAddress());
			return -1;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return 0;
	}
	
	
	/**
	 * @param b
	 * @param offset
	 * @param seq
	 * @return
	 * 
	 * A block with a unique sequence number in the first 4 bytes, rest data
	 * 
	 */
	private byte[] getBlock (byte[] data, int seq) {
		
		int offset = seq * ConInfo.UDPPACKETSIZE;
		
		byte[] result = new byte[ConInfo.UDPPACKETSIZE];
		
		byte[] bytedSeq = UtilMethods.intToBytes(seq); // Returns the 4 byte represntation of i
		
		for (int j = 0; j < 4; j++) {
			result[j] = bytedSeq[j];
		}
		
		for (int i = 4; i < ConInfo.UDPPACKETSIZE; i++) {
			result[i] = data[offset + i];
		}
		
		return result;
	}
	
}